
"""
django_auth_iam.models
~~~~~~~~~~~~~~~~~~~~~~~~

This module provides user and group models.

:copyright: (c) 2011  ViewWorld ApS
:license: GPLv3, see LICENSE for details.

"""

import bcrypt
import boto
from boto.exception import BotoServerError
from boto.sdb.db.model import Model
from boto.sdb.db.property import *
from boto.utils import Password

from .utils import encrypt_key, decrypt_key



class BCryptPassword(Password):

    def __init__(self, hash=None, hashfunc=None):
        if hash and not '$' in hash:
            raise ValueError('hash does not look like a proper bcrypt hash')
        elif hash is None:
            hash = ''
        super(BCryptPassword, self).__init__(hash, hashfunc)

    def set(self, value):
        """Set the password.

        The password will be hashed with the bcrypt algorithm before
        being stored.

        """
        if not isinstance(value, basestring):
            raise ValueError('value should be a string')
        if not value:
            self.str = ''
        else:
            self.str = bcrypt.hashpw(value, bcrypt.gensalt())

    def is_invalid(self):
        return self.str == ''

    def __eq__(self, other):
        """Returns ``True`` if the passwords match.

        The other password is hashed before comparing. The time taken
        is independent of the number of characters that match.

        """
        if not self.str or not other:
            return False
        hashed = str(self.str)
        other = bcrypt.hashpw(other, hashed)
        if len(hashed) != len(other):
            return False
        result = 0
        for x, y in zip(hashed, other):
            result |= ord(x) ^ ord(y)
        return result == 0

    def __ne__(self, other):
        if not self.str:
            return False
        return not (self == other)


class BCryptPasswordProperty(PasswordProperty):

    data_type = BCryptPassword
    type_name = 'BCryptPassword'

    def __get__(self, obj, objtype):
        return Property.__get__(self, obj, objtype)


class User(Model):

    username = StringProperty(required=True, unique=True)
    """Username for the user."""
    password = BCryptPasswordProperty()
    """Password for the user.

    The password is automatically hashed with :mod:`bcrypt` when it is
    stored. The password can be set and compared using standard
    operators.

    >>> user = User()
    >>> user.password = 'foo'
    >>> user.password == 'foo'
    True
    >>> user.password != 'bar'
    True
    >>> print user.password
    $2a$12$REbzH3KM5YdyAO6d8lGHSu2zeQZ2i4CzXpV4Y2HbIO7BYSCt9Plo.

    """
    access_key = StringProperty()
    """Access key to Amazon services."""
    enc_secret_key = StringProperty()
    """Encrypted secret key to Amazon services."""

    class AlreadyExist(Exception):
        def __init__(self, username):
            Exception.__init__(self, 'a user with the username "{0}" already '
                               'exist'.format(username))

    def get_secret_key(self, password):
        """Get the decrypted secret key.

        :param password: the unhashed password for the user.

        """
        return decrypt_key(self.enc_secret_key, password)

    @classmethod
    def get_by_username(cls, username):
        """Get a user by username. Returns ``None`` if the user is
        not found.

        """
        user = tuple(cls.find(username=username, limit=1))
        if len(user) == 0:
            return None
        return user[0]

    @classmethod
    def create(cls, username, password, force=False):
        """Create a new user. Raises :exc:`.User.AlreadyExist` if the
        user already exist in IAM.

        May also raise :class:`boto.exception.BotoServerError`.
        """
        if force:
            cls._delete_iam_user(username)
        cls._create_iam_user(username)
        user = cls.get_by_username(username)
        if user is None:
            user = cls()
        user.username = username
        user.password = password
        cls._create_access_key(user, password)
        user.put()
        return user

    @classmethod
    def _create_iam_user(cls, username):
        iam = boto.connect_iam()
        try:
            iam.create_user(username)
        except BotoServerError as e:
            if e.status == 409:
                raise cls.AlreadyExist(username)
            else:
                raise e

    @staticmethod
    def _create_access_key(user, password):
        iam = boto.connect_iam()
        response = iam.create_access_key(user.username)
        result = response['create_access_key_response']['create_access_key_result']
        access_key = result['access_key']['access_key_id']
        secret_key = result['access_key']['secret_access_key']
        enc_secret_key = encrypt_key(secret_key.encode('utf8'), password)
        user.access_key = access_key
        user.enc_secret_key = enc_secret_key

    def delete(self):
        """Delete the user from IAM and SimpleDB.

        May raise :exc:`boto.exception.BotoServerError`.

        """
        User._delete_iam_user(self.username)
        super(User, self).delete()

    @staticmethod
    def _delete_iam_user(username):
        try:
            iam = boto.connect_iam()
            response = iam.get_all_access_keys(username)
            key_ids = [k['access_key_id'] for k in
                       response['list_access_keys_response']
                       ['list_access_keys_result']['access_key_metadata']]
            for key in key_ids:
                iam.delete_access_key(key, username)
            iam.delete_user(username)
        except BotoServerError as e:
            if e.status != 404:
                raise e

class Group(Model):

    name = StringProperty(required=True, unique=True)
    """Group name."""

    class AlreadyExist(Exception):
        def __init__(self, name):
            Exception.__init__(self, 'a group with the name "{0}" already exist'
                               .format(name))

    @classmethod
    def get_by_name(cls, name):
        """Get a group by name. Returns ``None`` if the group is not found."""
        group = tuple(cls.find(name=name, limit=1))
        if len(group) == 0:
            return None
        return group[0]

    @classmethod
    def create(cls, name, force=False):
        """Create a new group. Raises :exc:`.Group.AlreadyExist` if
        the group already exist in IAM.

        May also raise :class:`boto.exception.BotoServerError`.

        """
        if force:
            cls._delete_iam_group(name)
        cls._create_iam_group(name)
        group = cls.get_by_name(name)
        if group is None:
            group = cls()
        group.put()
        return group

    @classmethod
    def _create_iam_group(cls, name):
        iam = boto.connect_iam()
        try:
            iam.create_group(name)
        except BotoServerError as e:
            if e.status == 409:
                raise cls.AlreadyExist(username)
            else:
                raise e

    def delete(self):
        """Delete the group from IAM and SimpleDB.

        May raise :exc:`boto.exception.BotoServerError`.

        """
        Group._delete_iam_group(self.name)
        super(Group, self).delete()

    @staticmethod
    def _delete_iam_group(name):
        try:
            iam = boto.connect_iam()
            iam.delete_group(name)
        except BotoServerError as e:
            if e.status != 404:
                raise e




"""
django_auth_iam.utils
~~~~~~~~~~~~~~~~~~~~~~~~

This module provides utility functions.

:copyright: (c) 2011  ViewWorld ApS
:license: GPLv3, see LICENSE for details.

"""

import hashlib
from Crypto.Cipher import AES


def encrypt_key(plain, password):
    if isinstance(plain, unicode):
        raise ValueError('unicode string is not allowed')
    encryption_key = hashlib.sha256(password).digest()
    encryptor = AES.new(encryption_key, AES.MODE_CBC)
    plain += ' ' * (16 - len(plain) % 16)
    cipher = encryptor.encrypt(plain)
    return cipher.encode('base64')

def decrypt_key(cipher, password):
    cipher = cipher.decode('base64')
    encryption_key = hashlib.sha256(password).digest()
    decryptor = AES.new(encryption_key, AES.MODE_CBC)
    plain = decryptor.decrypt(cipher).rstrip(' ')
    return plain



# -*- encoding: utf-8 -*-

import boto

from boto.sdb.db.model import Model
from django.test import TestCase

from . import utils
from .models import BCryptPassword, BCryptPasswordProperty

class TestEncryption(TestCase):
    def test_reversable(self):
        password = 'foobar123'
        key = 'AKIAIOSFODNN7EXAMPLE'
        enc_key = utils.encrypt_key(key, password)
        self.assertEqual(key, utils.decrypt_key(enc_key, password))
        # Test with unicode data
        key = u'AKIAIOSFODNN7EXHHE}O±J¸ZOXZÆL]}@¡ØÆLAMPLE'
        enc_key = utils.encrypt_key(key.encode('utf8'), password)
        self.assertEqual(key, utils.decrypt_key(enc_key, password).decode('utf8'))

    def test_types(self):
        password = 'foobar123'
        key = 'AKIAIOSFODNN7EXAMPLE'
        self.assertIsInstance(key, str)
        enc_key = utils.encrypt_key(key, password)
        self.assertIsInstance(enc_key, str)
        dec_key = utils.decrypt_key(enc_key, password)
        self.assertIsInstance(dec_key, str)


class TestBCryptPassword(TestCase):

    def test_init(self):
        p = BCryptPassword()
        self.assertEqual(str(p), '')
        self.assertEqual(len(p), 0)
        p = BCryptPassword('')
        self.assertEqual(str(p), '')
        self.assertEqual(len(p), 0)
        p = BCryptPassword('$2a$12$8Khdz89/84OwXYsN8zxCQ.AqYKMux9SagW1/sMEfWE9TLjN6dZ42m')
        self.assertNotEqual(str(p), '')
        self.assertNotEqual(len(p), 0)

    def test_set(self):
        p = BCryptPassword()
        p.set('')
        self.assertEqual(str(p), '')
        self.assertEqual(len(p), 0)
        p = BCryptPassword()
        p.set('test')
        self.assertNotEqual(str(p), '')
        self.assertNotEqual(len(p), 0)

    def test_compare(self):
        p = BCryptPassword()
        self.assertFalse(p == '')
        self.assertFalse(p != '')
        self.assertFalse(p == 'test')
        self.assertFalse(p == None)
        p.set('test')
        self.assertFalse(p == '')
        self.assertTrue(p != '')
        self.assertTrue(p == 'test')
        self.assertFalse(p == None)

    def test_invalid(self):
        p = BCryptPassword()
        self.assertTrue(p.is_invalid())
        p = BCryptPassword('')
        self.assertTrue(p.is_invalid())
        p = BCryptPassword('$2a$12$8Khdz89/84OwXYsN8zxCQ.AqYKMux9SagW1/sMEfWE9TLjN6dZ42m')
        self.assertFalse(p.is_invalid())

    def test_reinit(self):
        p = BCryptPassword()
        p.set('test')
        hash = str(p)
        p = BCryptPassword(hash)
        self.assertTrue(p == 'test')

    def test_badhash(self):
        with self.assertRaises(ValueError):
            p = BCryptPassword('badhash')


class TestModel(Model):
    __consistent__ = True

    p = BCryptPasswordProperty()

class TestBCryptPasswordProperty(TestCase):

    def setUp(self):
        sdb = boto.connect_sdb()
        sdb.create_domain('test_domain')

    def tearDown(self):
        sdb = boto.connect_sdb()
        sdb.delete_domain('test_domain')

    def test_settings(self):
        t = TestModel()
        self.assertEqual(str(t.p), '')
        self.assertFalse(t.p == '')
        self.assertTrue(t.p.is_invalid())
        t.p = 'test'
        self.assertNotEqual(str(t.p), '')
        self.assertTrue(t.p == 'test')
        self.assertFalse(t.p.is_invalid())
        
    def test_save(self):
        t = TestModel()
        t.p = 'test123'
        t = t.put()
        id = t.id
        t = TestModel.get_by_id(id)
        self.assertFalse(t.p.is_invalid())
        self.assertIsInstance(t.p.str, basestring)
        self.assertTrue(t.p == 'test123')



#!/usr/bin/env python

import os

from setuptools import setup
from distutils.cmd import Command

import django_auth_iam


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup(
    name='django-auth-iam',
    version=django_auth_iam.__version__,
    description='Django authentication backend using Amazon IAM',
    long_description=read('README.rst'),
    url='https://github.com/viewworld/django-auth-iam/',
    author='Michael Budde',
    author_email='mb@viewworld.dk',
    license='GPL v3',
    packages=['django_auth_iam'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python',
        'Topic :: Security',
        'Topic :: System :: Systems Administration :: Authentication/Directory',
    ],
    keywords=['django', 'amazon', 'authentication', 'auth'],
    install_requires=['boto', 'PyCrypto', 'py_bcrypt'],
)




"""
django_auth_iam.backends
~~~~~~~~~~~~~~~~~~~~~~~~

This module provides a backend for use with Django's authentication
framework.

:copyright: (c) 2011  ViewWorld Aps
:license: GPLv3, see LICENSE for details.

"""

from django.conf import settings
from django.contrib.auth.models import User
from django.contrib.auth.backends import ModelBackend
from django.utils.importlib import import_module

import logging
logger = logging.getLogger('django_auth_iam')


IAM_USER_CLASS = getattr(settings, 'IAM_USER_CLASS', 'django_auth_iam.models.User')


class AmazonIAMBackend(ModelBackend):
    """
    Authenticates against Amazon's IAM service.
    """

    supports_object_permissions = False
    supports_anonymous_user = False
    supports_inactive_user = False

    def __init__(self):
        self.user_class = self._load_user_class()

    def _load_user_class(self):
        module, sep, attr = IAM_USER_CLASS.rpartition('.')
        try:
            mod = import_module(module)
        except ImportError, e:
            raise ImproperlyConfigured('Error importing user class module '
                                       '{0}: "{1}"'.format(path, e))
        except ValueError, e:
            raise ImproperlyConfigured('Error importing user class module. '
                                       'Is IAM_USER_CLASS a string?')
        try:
            cls = getattr(mod, attr)
        except AttributeError:
            raise ImproperlyConfigured('Module "{0}" does not define a '
                                       '"{1}" user class'.format(module, attr))
        return cls

    def authenticate(self, username=None, password=None):
        iamuser = self.user_class.get_by_username(username)
        if iamuser is None or iamuser.password != password:
            logger.info('Authentication FAILED for user `{0}`'.format(username))
            return None
        secret_key = iamuser.get_secret_key(password)
        user, created = User.objects.get_or_create(username=username)
        user.aws_credentials = (iamuser.access_key, secret_key)
        user.iam_user = iamuser
        logger.info('Authentication SUCCEEDED for user `{0}`'.format(username))
        return user
